<?php
// $Id: millennium.d6.inc,v 1.1.2.2 2010/05/06 16:58:24 janusman Exp $

/**
 * @file
 *   Functions from Drupal 6.
 *
 *   Make our life easily on backporting.
 */

if (!function_exists('drupal_alter')) {
  function drupal_alter($type, &$data) {
    // PHP's func_get_args() always returns copies of params, not references, so
    // drupal_alter() can only manipulate data that comes in via the required first
    // param. For the edge case functions that must pass in an arbitrary number of
    // alterable parameters (hook_form_alter() being the best example), an array of
    // those params can be placed in the __drupal_alter_by_ref key of the $data
    // array. This is somewhat ugly, but is an unavoidable consequence of a flexible
    // drupal_alter() function, and the limitations of func_get_args().
    // @todo: Remove this in Drupal 7.
    if (is_array($data) && isset($data['__drupal_alter_by_ref'])) {
      $by_ref_parameters = $data['__drupal_alter_by_ref'];
      unset($data['__drupal_alter_by_ref']);
    }

    // Hang onto a reference to the data array so that it isn't blown away later.
    // Also, merge in any parameters that need to be passed by reference.
    $args = array(&$data);
    if (isset($by_ref_parameters)) {
      $args = array_merge($args, $by_ref_parameters);
    }

    // Now, use func_get_args() to pull in any additional parameters passed into
    // the drupal_alter() call.
    $additional_args = func_get_args();
    array_shift($additional_args);
    array_shift($additional_args);
    $args = array_merge($args, $additional_args);

    foreach (module_implements($type .'_alter') as $module) {
      $function = $module .'_'. $type .'_alter';
      call_user_func_array($function, $args);
    }
  }
}

if (!function_exists('db_placeholders')) {
  function db_placeholders($arguments, $type = 'int') {
    $placeholder = db_type_placeholder($type);
    return implode(',', array_fill(0, count($arguments), $placeholder));
  }
}
if (!function_exists('db_type_placeholder')) {
  function db_type_placeholder($type) {
    switch ($type) {
      case 'varchar':
      case 'char':
      case 'text':
      case 'datetime':
        return "'%s'";

      case 'numeric':
        // Numeric values are arbitrary precision numbers.  Syntacically, numerics
        // should be specified directly in SQL. However, without single quotes
        // the %s placeholder does not protect against non-numeric characters such
        // as spaces which would expose us to SQL injection.
        return '%n';

      case 'serial':
      case 'int':
        return '%d';

      case 'float':
        return '%f';

      case 'blob':
        return '%b';
    }

    // There is no safe value to return here, so return something that
    // will cause the query to fail.
    return 'unsupported type '. $type .'for db_type_placeholder';
  }
}

if (!function_exists('drupal_write_record')) {
  /**
   * Backport of drupal_write_record(). Depends on contrib Schema API module.
   */
  function drupal_write_record($table, &$object, $update = array()) {
    // Standardize $update to an array.
    if (is_string($update)) {
      $update = array($update);
    }

    $schema = drupal_get_schema($table);
    if (empty($schema)) {
      return FALSE;
    }

    // Convert to an object if needed.
    if (is_array($object)) {
      $object = (object) $object;
      $array = TRUE;
    }
    else {
      $array = FALSE;
    }

    $fields = $defs = $values = $serials = $placeholders = array();

    // Go through our schema, build SQL, and when inserting, fill in defaults for
    // fields that are not set.
    foreach ($schema['fields'] as $field => $info) {
      // Special case -- skip serial types if we are updating.
      if ($info['type'] == 'serial' && count($update)) {
        continue;
      }

      // For inserts, populate defaults from Schema if not already provided
      if (!isset($object->$field) && !count($update) && isset($info['default'])) {
        $object->$field = $info['default'];
      }

      // Track serial fields so we can helpfully populate them after the query.
      if ($info['type'] == 'serial') {
        $serials[] = $field;
        // Ignore values for serials when inserting data. Unsupported.
        unset($object->$field);
      }

      // Build arrays for the fields, placeholders, and values in our query.
      if (isset($object->$field)) {
        $fields[] = $field;
        $placeholders[] = db_type_placeholder($info['type']);

        if (empty($info['serialize'])) {
          $values[] = $object->$field;
        }
        else {
          $values[] = serialize($object->$field);
        }
      }
    }

    // Build the SQL.
    $query = '';
    if (!count($update)) {
      $query = "INSERT INTO {". $table ."} (". implode(', ', $fields) .') VALUES ('. implode(', ', $placeholders) .')';
      $return = SAVED_NEW;
    }
    else {
      $query = '';
      foreach ($fields as $id => $field) {
        if ($query) {
          $query .= ', ';
        }
        $query .= $field .' = '. $placeholders[$id];
      }

      foreach ($update as $key){
        $conditions[] = "$key = ". db_type_placeholder($schema['fields'][$key]['type']);
        $values[] = $object->$key;
      }

      $query = "UPDATE {". $table ."} SET $query WHERE ". implode(' AND ', $conditions);
      $return = SAVED_UPDATED;
    }

    // Execute the SQL.
    if (db_query($query, $values)) {
      if ($serials) {
        // Get last insert ids and fill them in.
        foreach ($serials as $field) {
          $object->$field = db_last_insert_id($table, $field);
        }
      }
    }
    else {
      $return = FALSE;
    }

    // If we began with an array, convert back so we don't surprise the caller.
    if ($array) {
      $object = (array) $object;
    }

    return $return;
  }
}

if (!function_exists('language_default')) {
  function language_default($property = NULL) {
    $language = variable_get('language_default', (object) array('language' => 'en', 'name' => 'English', 'native' => 'English', 'direction' => 0, 'enabled' => 1, 'plurals' => 0, 'formula' => '', 'domain' => '', 'prefix' => '', 'weight' => 0, 'javascript' => ''));
    return $property ? $language->$property : $language;
  }
}
